import * as plugins from './plugins.js';
import * as interfaces from './env.js';
import { logger } from './logging.js';
import { ServiceWorker } from './classes.serviceworker.js';

export class CacheManager {
  public losslessServiceWorkerRef: ServiceWorker;

  public usedCacheNames = {
    runtimeCacheName: 'runtime'
  };

  constructor(losslessServiceWorkerRefArg: ServiceWorker) {
    this.losslessServiceWorkerRef = losslessServiceWorkerRefArg;
    this._setupCache();
  }

  private _setupCache = () => {
    const createMatchRequest = (requestArg: Request) => {
      // Create a matchRequest.
      let matchRequest: Request;
      if (
        requestArg.url.startsWith(
          this.losslessServiceWorkerRef.serviceWindowRef.location.origin
        )
      ) {
        // internal request
        matchRequest = requestArg;
      } else {
        // For external requests, create a new Request with CORS settings.
        matchRequest = new Request(requestArg.url, {
          method: requestArg.method,
          headers: requestArg.headers,
          mode: 'cors',
          credentials: 'same-origin',
          redirect: 'follow'
        });
      }
      return matchRequest;
    };

    /**
     * Creates a 500 response
     */
    const create500Response = async (requestArg: Request, responseArg: Response) => {
      return new Response(
        `
        <html>
          <head>
            <style>
              .note {
                padding: 10px;
                color: #fff;
                background: #000;
                border-bottom: 1px solid #e4002b;
                margin-bottom: 20px;
              }
            </style>
          </head>
          <body>
          
          <div class="note">
            <strong>serviceworker running, but status 500</strong><br>
          </div>
          serviceworker is unable to fetch this request<br>
          Here is some info about the request/response pair:<br>
          <br>
          requestUrl: ${requestArg.url}<br>
          responseType: ${responseArg.type}<br>
          responseBody: ${await responseArg.clone().text()}<br>
          </body>
        </html>
      `,
        {
          headers: {
            "Content-Type": "text/html"
          },
          status: 500
        }
      );
    };

    // A list of local resources we always want to be cached.
    this.losslessServiceWorkerRef.serviceWindowRef.addEventListener('fetch', async (fetchEventArg: any) => {
      // Block scopes we don't want the serviceworker to handle.
      const originalRequest: Request = fetchEventArg.request;
      const parsedUrl = new URL(originalRequest.url);
      if (
        parsedUrl.hostname.includes('paddle.com') ||
        parsedUrl.hostname.includes('paypal.com') ||
        parsedUrl.hostname.includes('reception.lossless.one') ||
        parsedUrl.pathname.startsWith('/socket.io') ||
        originalRequest.url.startsWith('https://umami.')
      ) {
        logger.log('note', `serviceworker not active for ${parsedUrl.toString()}`);
        return;
      }

      // Continue for the rest.
      const done = plugins.smartpromise.defer<Response>();
      fetchEventArg.respondWith(done.promise);

      if (
        (originalRequest.method === 'GET' &&
          (originalRequest.url.startsWith(this.losslessServiceWorkerRef.serviceWindowRef.location.origin) &&
            !originalRequest.url.includes('/api/') &&
            !originalRequest.url.includes('smartserve/reloadcheck'))) ||
        originalRequest.url.includes('https://assetbroker.') ||
        originalRequest.url.includes('https://unpkg.com') ||
        originalRequest.url.includes('https://fonts.googleapis.com') ||
        originalRequest.url.includes('https://fonts.gstatic.com')
      ) {
        // Check for updates asynchronously.
        this.losslessServiceWorkerRef.updateManager.checkUpdate(this);

        // Handle local or approved remote requests.
        const matchRequest = createMatchRequest(originalRequest);
        const cachedResponse = await caches.match(matchRequest);
        if (cachedResponse) {
          logger.log('ok', `CACHED: found cached response for ${matchRequest.url}`);
          done.resolve(cachedResponse);
          return;
        }

        // No cached response found; try to fetch and cache.
        logger.log('info', `NOTYETCACHED: trying to cache ${matchRequest.url}`);
        const newResponse: Response = await fetch(matchRequest).catch(async err => {
          return await create500Response(matchRequest, new Response(err.message));
        });

        // If status > 299 or opaque response, don't cache.
        if (newResponse.status > 299 || newResponse.type === 'opaque') {
          logger.log(
            'error',
            `NOTCACHED: can't cache response for ${matchRequest.url} due to status ${
              newResponse.status
            } and type ${newResponse.type}`
          );
          done.resolve(await create500Response(matchRequest, newResponse));
        } else {
          const cache = await caches.open(this.usedCacheNames.runtimeCacheName);
          const responseToPutToCache = newResponse.clone();

          // Create new headers preserving all except caching-related headers.
          const headers = new Headers();
          responseToPutToCache.headers.forEach((value, key) => {
            if (![
              'Cache-Control',
              'cache-control',
              'Expires',
              'expires',
              'Pragma',
              'pragma'
            ].includes(key)) {
              headers.set(key, value);
            }
          });

          // Ensure CORS headers are present in cached response.
          if (!headers.has('Access-Control-Allow-Origin')) {
            headers.set('Access-Control-Allow-Origin', '*');
          }
          if (!headers.has('Access-Control-Allow-Methods')) {
            headers.set('Access-Control-Allow-Methods', 'GET, POST, PUT, DELETE, OPTIONS');
          }
          if (!headers.has('Access-Control-Allow-Headers')) {
            headers.set('Access-Control-Allow-Headers', 'Content-Type');
          }
          // Prevent browser caching while allowing service worker caching.
          headers.set('Cache-Control', 'no-store, no-cache, must-revalidate, proxy-revalidate');
          headers.set('Pragma', 'no-cache');
          headers.set('Expires', '0');
          headers.set('Surrogate-Control', 'no-store');

          // IMPORTANT: Read the entire response body as a blob so that
          // Safari does not have issues with a locked stream when caching.
          const bodyBlob = await responseToPutToCache.blob();
          const newCachedResponse = new Response(bodyBlob, {
            status: responseToPutToCache.status,
            statusText: responseToPutToCache.statusText,
            headers
          });
          await cache.put(matchRequest, newCachedResponse);
          logger.log('ok', `NOWCACHED: cached response for ${matchRequest.url} for subsequent requests!`);
          done.resolve(newResponse);
        }
      } else {
        // For remote requests that don't qualify for caching.
        logger.log(
          'ok',
          `NOTCACHED: not caching any responses for ${originalRequest.url}. Fetching from origin now...`
        );
        done.resolve(
          await fetch(originalRequest).catch(async err => {
            return await create500Response(originalRequest, new Response(err.message));
          })
        );
      }
    });
  }

  /**
   * Cleans all caches.
   * Should only be run when running a new service worker.
   * @param reasonArg
   */
  public cleanCaches = async (reasonArg = 'no reason given') => {
    logger.log('info', `MAJOR CACHEEVENT: cleaning caches now! Reason: ${reasonArg}`);
    const cacheNames = await caches.keys();

    const deletePromises = cacheNames.map(cacheToDelete => {
      const deletePromise = caches.delete(cacheToDelete);
      deletePromise.then(() => {
        logger.log('ok', `Deleted cache ${cacheToDelete}`);
      });
      return deletePromise;
    });
    await Promise.all(deletePromises);
  }

  /**
   * Revalidates the runtime cache.
   */
  public async revalidateCache() {
    const runtimeCache = await caches.open(this.usedCacheNames.runtimeCacheName);
    const cacheKeys = await runtimeCache.keys();
    for (const requestArg of cacheKeys) {
      // Get the cached response.
      const cachedResponse = runtimeCache.match(requestArg);

      // Fetch a new response for comparison.
      const clonedRequest = requestArg.clone();
      const response = await plugins.smartpromise.timeoutWrap(fetch(clonedRequest), 5000); // Increased timeout for better mobile compatibility
      if (response && response.status >= 200 && response.status < 300) {
        await runtimeCache.delete(requestArg);
        await runtimeCache.put(requestArg, response);
      }
    }
  }
}